/* Sottoscrittore.java */

import java.util.Vector;

/**
 * @author Fosco
 * 
 * Questa classe implementa il generico consumatore
 */
public class Sottoscrittore extends Thread {
  private Canale canale;

  private int identificatore = -1;

  private Vector codaMessaggi = new Vector();

  public Sottoscrittore(Canale c, int id) {
    super("S" + id);
    canale = c;
    identificatore = id;
  }

  public void run() {
    try {
      sleep((int) (Math.random() * 3000) + 2000);

      System.out.println("\nSottoscrittore (S" + identificatore
          + "): inizio e sottoscrizione...");

      canale.addSubscriber(this);

      System.out.println("Il sottoscrittore S" + identificatore
          + " e' pronto a ricevere " + Editore.MAX_PUBBLICAZIONI + " messaggi");

      /*
       * Per semplicita' il singolo sottoscrittore fa un numero di letture molto
       * minore di tutte le scritture effettuate dagli editori (pari al numero
       * degli editori * MAX_PUBBLICAZIONI). Come si potrebbe gestire in modo
       * piu' corretto la terminazione di tutti i processi?
       * 
       * Si provi a svolgere le proposta di estensione.
       *  
       */
      for (int i = 0; i < Editore.MAX_PUBBLICAZIONI; i++) {
        System.out.println("S" + identificatore + ", lettura " + i + ": "
            + read());
      }

      canale.removeSubscriber(this);

      System.out.println("\nS" + identificatore
          + ": de-sottoscrizione e terminazione.");
    }
    catch (Exception e) {
      System.err.println("Si e' verificato un errore, il seguente: " + e);
      e.printStackTrace(System.err);
    }

  }

  /**
   * Chiamata sospensiva in caso la coda dei messaggi sia vuota. NOTA: come
   * effetto collaterale la read consuma il messaggio dalla coda dei messaggi.
   * 
   * @return il messaggio ricevuto, la stringa "-1" come EOF e null se ci sono
   *         problemi
   */
  private synchronized String read() throws InterruptedException {
    /* Perche' basta l'if? Quanti processi eseguono questo metodo? */
    if (codaMessaggi.size() == 0)
      wait();
    String firstReceivedMessage = (String) codaMessaggi.lastElement();
    codaMessaggi.remove(codaMessaggi.size() - 1);
    return firstReceivedMessage;
  }

  public synchronized void push(String msg) {
    codaMessaggi.insertElementAt(msg, 0);
    /*
     * NOTA: perche' in questo caso il metodo push DEVE essere synchronized?
     * Quanti thread accedono simultaneamente a questo metodo?
     * 
     * Soluzione alternativa a quella nell'esercitazione svolta. Spostiamo la
     * notify direttamente all'interno di questo metodo.
     * 
     * Perche' basta notify e non notifyAll? Quanti thread possono essere
     * bloccati?
     */
    notify();
  }

}